import logging
import os
import requests

from kube_hunter.conf import config


class HTTPDispatcher(object):
    def dispatch(self, report):
        logging.debug('Dispatching report via http')
        dispatch_method = os.environ.get(
            'KUBEHUNTER_HTTP_DISPATCH_METHOD',
            'POST'
        ).upper()
        dispatch_url = os.environ.get(
            'KUBEHUNTER_HTTP_DISPATCH_URL',
            'https://localhost/'
        )
        try:
            r = requests.request(
                dispatch_method,
                dispatch_url,
                json=report,
                headers={'Content-Type': 'application/json'}
            )
            r.raise_for_status()
            logging.info('\nReport was dispatched to: {url}'.format(url=dispatch_url))
            logging.debug(
                "\tResponse Code: {status}\n\tResponse Data:\n{data}".format(
                    status=r.status_code,
                    data=r.text
                )
            )
        except requests.HTTPError as e:
            # specific http exceptions
            logging.exception(
                "\nCould not dispatch report using HTTP {method} to {url}\nResponse Code: {status}".format(
                    status=r.status_code,
                    url=dispatch_url,
                    method=dispatch_method
                )
            )
        except Exception as e:
            # default all exceptions
            logging.exception("\nCould not dispatch report using HTTP {method} to {url}".format(
                method=dispatch_method,
                url=dispatch_url
            ))

class STDOUTDispatcher(object):
    def dispatch(self, report):
        logging.debug('Dispatching report via stdout')
        if config.report == "plain":
            logging.info("\n{div}".format(div="-" * 10))
        print(report)



import logging

import requests

from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Vulnerability, Event, OpenPortEvent
from kube_hunter.core.types import ActiveHunter, Hunter, KubernetesCluster, \
    InformationDisclosure, RemoteCodeExec, UnauthenticatedAccess, AccessRisk


""" Vulnerabilities """
class EtcdRemoteWriteAccessEvent(Vulnerability, Event):
    """Remote write access might grant an attacker full control over the kubernetes cluster"""

    def __init__(self, write_res):
        Vulnerability.__init__(self, KubernetesCluster, name="Etcd Remote Write Access Event", category=RemoteCodeExec, vid="KHV031")
        self.evidence = write_res


class EtcdRemoteReadAccessEvent(Vulnerability, Event):
    """Remote read access might expose to an attacker cluster's possible exploits, secrets and more."""

    def __init__(self, keys):
        Vulnerability.__init__(self, KubernetesCluster,  name="Etcd Remote Read Access Event", category=AccessRisk, vid="KHV032")
        self.evidence = keys


class EtcdRemoteVersionDisclosureEvent(Vulnerability, Event):
    """Remote version disclosure might give an attacker a valuable data to attack a cluster"""

    def __init__(self, version):

        Vulnerability.__init__(self, KubernetesCluster, name="Etcd Remote version disclosure",
                               category=InformationDisclosure, vid="KHV033")
        self.evidence = version


class EtcdAccessEnabledWithoutAuthEvent(Vulnerability, Event):
    """Etcd is accessible using HTTP (without authorization and authentication), it would allow a potential attacker to
     gain access to the etcd"""

    def __init__(self, version):
        Vulnerability.__init__(self, KubernetesCluster,  name="Etcd is accessible using insecure connection (HTTP)",
                               category=UnauthenticatedAccess, vid="KHV034")
        self.evidence = version


# Active Hunter
@handler.subscribe(OpenPortEvent, predicate=lambda p: p.port == 2379)
class EtcdRemoteAccessActive(ActiveHunter):
    """Etcd Remote Access
    Checks for remote write access to etcd- will attempt to add a new key to the etcd DB"""

    def __init__(self, event):
        self.event = event
        self.write_evidence = ''

    def db_keys_write_access(self):
        logging.debug("Active hunter is attempting to write keys remotely on host " + self.event.host)
        data = {
            'value': 'remotely written data'
        }
        try:
            r = requests.post("{protocol}://{host}:{port}/v2/keys/message".format(host=self.event.host, port=2379,
                                                                                  protocol=self.protocol), data=data)
            self.write_evidence = r.content if r.status_code == 200 and r.content != '' else False
            return self.write_evidence
        except requests.exceptions.ConnectionError:
            return False

    def execute(self):
        if self.db_keys_write_access():
            self.publish_event(EtcdRemoteWriteAccessEvent(self.write_evidence))


# Passive Hunter
@handler.subscribe(OpenPortEvent, predicate=lambda p: p.port == 2379)
class EtcdRemoteAccess(Hunter):
    """Etcd Remote Access
    Checks for remote availability of etcd, its version, and read access to the DB
    """

    def __init__(self, event):
        self.event = event
        self.version_evidence = ''
        self.keys_evidence = ''
        self.protocol = 'https'

    def db_keys_disclosure(self):
        logging.debug(self.event.host + " Passive hunter is attempting to read etcd keys remotely")
        try:
            r = requests.get(
                "{protocol}://{host}:{port}/v2/keys".format(protocol=self.protocol, host=self.event.host, port=2379),
                verify=False)
            self.keys_evidence = r.content if r.status_code == 200 and r.content != '' else False
            return self.keys_evidence
        except requests.exceptions.ConnectionError:
            return False

    def version_disclosure(self):
        logging.debug(self.event.host + " Passive hunter is attempting to check etcd version remotely")
        try:
            r = requests.get(
                "{protocol}://{host}:{port}/version".format(protocol=self.protocol, host=self.event.host, port=2379),
                verify=False)
            self.version_evidence = r.content if r.status_code == 200 and r.content != '' else False
            return self.version_evidence
        except requests.exceptions.ConnectionError:
            return False

    def insecure_access(self):
        logging.debug(self.event.host + " Passive hunter is attempting to access etcd insecurely")
        try:
            r = requests.get("http://{host}:{port}/version".format(host=self.event.host, port=2379), verify=False)
            return r.content if r.status_code == 200 and r.content != '' else False
        except requests.exceptions.ConnectionError:
            return False

    def execute(self):
        if self.insecure_access():  # make a decision between http and https protocol
            self.protocol = 'http'
        if self.version_disclosure():
            self.publish_event(EtcdRemoteVersionDisclosureEvent(self.version_evidence))
            if self.protocol == 'http':
                self.publish_event(EtcdAccessEnabledWithoutAuthEvent(self.version_evidence))
            if self.db_keys_disclosure():
                self.publish_event(EtcdRemoteReadAccessEvent(self.keys_evidence))



import json
import logging

import requests

from kube_hunter.modules.hunting.kubelet import ExposedRunHandler
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Event, Vulnerability
from kube_hunter.core.types import Hunter, ActiveHunter, IdentityTheft, Azure


class AzureSpnExposure(Vulnerability, Event):
    """The SPN is exposed, potentially allowing an attacker to gain access to the Azure subscription"""
    def __init__(self, container):
        Vulnerability.__init__(self, Azure, "Azure SPN Exposure", category=IdentityTheft, vid="KHV004")
        self.container = container

@handler.subscribe(ExposedRunHandler, predicate=lambda x: x.cloud=="Azure")
class AzureSpnHunter(Hunter):
    """AKS Hunting
    Hunting Azure cluster deployments using specific known configurations
    """
    def __init__(self, event):
        self.event = event
        self.base_url = "https://{}:{}".format(self.event.host, self.event.port)

    # getting a container that has access to the azure.json file
    def get_key_container(self):
        logging.debug("Passive Hunter is attempting to find container with access to azure.json file")
        raw_pods = requests.get(self.base_url + "/pods", verify=False).text
        if "items" in raw_pods:
            pods_data = json.loads(raw_pods)["items"]
            for pod_data in pods_data:
                for container in pod_data["spec"]["containers"]:
                    for mount in container["volumeMounts"]:
                        path = mount["mountPath"]
                        if '/etc/kubernetes/azure.json'.startswith(path):
                            return {
                                "name": container["name"],
                                "pod": pod_data["metadata"]["name"],
                                "namespace": pod_data["metadata"]["namespace"]
                            }

    def execute(self):
        container = self.get_key_container()
        if container:
            self.publish_event(AzureSpnExposure(container=container))

""" Active Hunting """
@handler.subscribe(AzureSpnExposure)
class ProveAzureSpnExposure(ActiveHunter):
    """Azure SPN Hunter
    Gets the azure subscription file on the host by executing inside a container
    """
    def __init__(self, event):
        self.event = event
        self.base_url = "https://{}:{}".format(self.event.host, self.event.port)

    def run(self, command, container):
        run_url = "{base}/run/{pod_namespace}/{pod_id}/{container_name}".format(
            base=self.base_url,
            pod_namespace=container["namespace"],
            pod_id=container["pod"],
            container_name=container["name"]
        )
        return requests.post(run_url, verify=False, params={'cmd': command}).text

    def execute(self):
        raw_output = self.run("cat /etc/kubernetes/azure.json", container=self.event.container)
        if "subscriptionId" in raw_output:
            subscription = json.loads(raw_output)
            self.event.subscriptionId = subscription["subscriptionId"]
            self.event.aadClientId = subscription["aadClientId"]
            self.event.aadClientSecret = subscription["aadClientSecret"]
            self.event.tenantId = subscription["tenantId"]
            self.event.evidence = "subscription: {}".format(self.event.subscriptionId)



import logging

from socket import socket

from kube_hunter.core.types import Discovery
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import NewHostEvent, OpenPortEvent


default_ports = [8001, 8080, 10250, 10255, 30000, 443, 6443, 2379]

@handler.subscribe(NewHostEvent)
class PortDiscovery(Discovery):
    """Port Scanning
    Scans Kubernetes known ports to determine open endpoints for discovery
    """
    def __init__(self, event):
        self.event = event
        self.host = event.host
        self.port = event.port

    def execute(self):
        logging.debug("host {0} try ports: {1}".format(self.host, default_ports))
        for single_port in default_ports:
            if self.test_connection(self.host, single_port):
                logging.debug("Reachable port found: {0}".format(single_port))
                self.publish_event(OpenPortEvent(port=single_port))

    @staticmethod
    def test_connection(host, port):
        s = socket()
        s.settimeout(1.5)
        try:
            success = s.connect_ex((str(host), port))
            if success == 0:
                return True
        except: pass
        finally: s.close()
        return False



import time
import requests_mock

from kube_hunter.core.events import handler
from kube_hunter.core.events.types import K8sVersionDisclosure
from kube_hunter.modules.hunting.cves import K8sClusterCveHunter, ServerApiVersionEndPointAccessPE, ServerApiVersionEndPointAccessDos, CveUtils

cve_counter = 0

def test_K8sCveHunter():
    global cve_counter
    # because the hunter unregisters itself, we manually remove this option, so we can test it
    K8sClusterCveHunter.__new__ = lambda self, cls: object.__new__(self)

    e = K8sVersionDisclosure(version="1.10.1", from_endpoint="/version")
    h = K8sClusterCveHunter(e)
    h.execute()

    time.sleep(0.01)
    assert cve_counter == 2
    cve_counter = 0

    # test patched version
    e = K8sVersionDisclosure(version="v1.13.6-gke.13", from_endpoint="/version")
    h = K8sClusterCveHunter(e)
    h.execute()

    time.sleep(0.01)
    assert cve_counter == 0
    cve_counter = 0


@handler.subscribe(ServerApiVersionEndPointAccessPE)
class test_CVE_2018_1002105(object):
    def __init__(self, event):
        global cve_counter
        cve_counter += 1

@handler.subscribe(ServerApiVersionEndPointAccessDos)
class test_CVE_2019_1002100(object):
    def __init__(self, event):
        global cve_counter
        cve_counter += 1

class test_CveUtils(object):
    def test_is_downstream():
        test_cases = (
            ('1', False),
            ('1.2', False),
            ('1.2-3', True),
            ('1.2-r3', True),
            ('1.2+3', True),
            ('1.2~3', True),
            ('1.2+a3f5cb2', True),
            ('1.2-9287543', True),
            ('v1', False),
            ('v1.2', False),
            ('v1.2-3', True),
            ('v1.2-r3', True),
            ('v1.2+3', True),
            ('v1.2~3', True),
            ('v1.2+a3f5cb2', True),
            ('v1.2-9287543', True),
            ('v1.13.9-gke.3', True)
        )

        for version, expected in test_cases:
            actual = CveUtils.is_downstream_version(version)
            assert actual == expected

    def test_ignore_downstream():
        test_cases = (
            ('v2.2-abcd', ['v1.1', 'v2.3'], False),
            ('v2.2-abcd', ['v1.1', 'v2.2'], False),
            ('v1.13.9-gke.3', ['v1.14.8'], False)
        )

        for check_version, fix_versions, expected in test_cases:
            actual = CveUtils.is_vulnerable(check_version, fix_versions, True)
            assert actual == expected



from kube_hunter.modules.report.json import JSONReporter
from kube_hunter.modules.report.yaml import YAMLReporter
from kube_hunter.modules.report.plain import PlainReporter
from kube_hunter.modules.report.dispatchers import \
    STDOUTDispatcher, HTTPDispatcher

import logging

reporters = {
    'yaml': YAMLReporter,
    'json': JSONReporter,
    'plain': PlainReporter
}

dispatchers = {
    'stdout': STDOUTDispatcher,
    'http': HTTPDispatcher
}


def get_reporter(name):
    try:
        return reporters[name.lower()]()
    except KeyError:
        logging.warning('Unknown reporter selected, using plain')
        return reporters['plain']()


def get_dispatcher(name):
    try:
        return dispatchers[name.lower()]()
    except KeyError:
        logging.warning('Unknown dispatcher selected, using stdout')
        return dispatchers['stdout']()



import logging
import json
import requests
import uuid
import copy

from kube_hunter.modules.discovery.apiserver import ApiServer
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Vulnerability, Event, K8sVersionDisclosure
from kube_hunter.core.types import Hunter, ActiveHunter, KubernetesCluster
from kube_hunter.core.types import RemoteCodeExec, AccessRisk, InformationDisclosure, UnauthenticatedAccess


class ServerApiAccess(Vulnerability, Event):
    """ The API Server port is accessible. Depending on your RBAC settings this could expose access to or control of your cluster. """

    def __init__(self, evidence, using_token):
        if using_token:
            name = "Access to API using service account token"
            category = InformationDisclosure
        else:
            name = "Unauthenticated access to API"
            category = UnauthenticatedAccess
        Vulnerability.__init__(self, KubernetesCluster, name=name, category=category, vid="KHV005")
        self.evidence = evidence

class ServerApiHTTPAccess(Vulnerability, Event):
    """ The API Server port is accessible over HTTP, and therefore unencrypted. Depending on your RBAC settings this could expose access to or control of your cluster. """

    def __init__(self, evidence):
        name = "Insecure (HTTP) access to API"
        category = UnauthenticatedAccess
        Vulnerability.__init__(self, KubernetesCluster, name=name, category=category, vid="KHV006")
        self.evidence = evidence

class ApiInfoDisclosure(Vulnerability, Event):
    def __init__(self, evidence, using_token, name):
        if using_token:
            name +=" using service account token"
        else:
            name +=" as anonymous user"
        Vulnerability.__init__(self, KubernetesCluster, name=name, category=InformationDisclosure, vid="KHV007")
        self.evidence = evidence

class ListPodsAndNamespaces(ApiInfoDisclosure):
    """ Accessing pods might give an attacker valuable information"""

    def __init__(self, evidence, using_token):
        ApiInfoDisclosure.__init__(self, evidence, using_token, "Listing pods")


class ListNamespaces(ApiInfoDisclosure):
    """ Accessing namespaces might give an attacker valuable information """

    def __init__(self, evidence, using_token):
        ApiInfoDisclosure.__init__(self, evidence, using_token, "Listing namespaces")


class ListRoles(ApiInfoDisclosure):
    """ Accessing roles might give an attacker valuable information """

    def __init__(self, evidence, using_token):
        ApiInfoDisclosure.__init__(self, evidence, using_token, "Listing roles")


class ListClusterRoles(ApiInfoDisclosure):
    """ Accessing cluster roles might give an attacker valuable information """

    def __init__(self, evidence, using_token):
        ApiInfoDisclosure.__init__(self, evidence, using_token, "Listing cluster roles")


class CreateANamespace(Vulnerability, Event):

    """ Creating a namespace might give an attacker an area with default (exploitable) permissions to run pods in.
    """
    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Created a namespace",
                               category=AccessRisk)
        self.evidence = evidence


class DeleteANamespace(Vulnerability, Event):

    """ Deleting a namespace might give an attacker the option to affect application behavior """
    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Delete a namespace",
                               category=AccessRisk)
        self.evidence = evidence


class CreateARole(Vulnerability, Event):
    """ Creating a role might give an attacker the option to harm the normal behavior of newly created pods
     within the specified namespaces.
    """

    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Created a role",
                               category=AccessRisk)
        self.evidence = evidence


class CreateAClusterRole(Vulnerability, Event):
    """ Creating a cluster role might give an attacker the option to harm the normal behavior of newly created pods
     across the whole cluster
    """

    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Created a cluster role",
                               category=AccessRisk)
        self.evidence = evidence


class PatchARole(Vulnerability, Event):
    """ Patching a role might give an attacker the option to create new pods with custom roles within the
    specific role's namespace scope
    """

    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Patched a role",
                               category=AccessRisk)
        self.evidence = evidence


class PatchAClusterRole(Vulnerability, Event):
    """ Patching a cluster role might give an attacker the option to create new pods with custom roles within the whole
    cluster scope.
    """

    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Patched a cluster role",
                               category=AccessRisk)
        self.evidence = evidence


class DeleteARole(Vulnerability, Event):
    """ Deleting a role might allow an attacker to affect access to resources in the namespace"""

    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Deleted a role",
                               category=AccessRisk)
        self.evidence = evidence


class DeleteAClusterRole(Vulnerability, Event):
    """ Deleting a cluster role might allow an attacker to affect access to resources in the cluster"""

    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Deleted a cluster role",
                               category=AccessRisk)
        self.evidence = evidence


class CreateAPod(Vulnerability, Event):
    """ Creating a new pod allows an attacker to run custom code"""

    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Created A Pod",
                               category=AccessRisk)
        self.evidence = evidence


class CreateAPrivilegedPod(Vulnerability, Event):
    """ Creating a new PRIVILEGED pod would gain an attacker FULL CONTROL over the cluster"""

    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Created A PRIVILEGED Pod",
                               category=AccessRisk)
        self.evidence = evidence


class PatchAPod(Vulnerability, Event):
    """ Patching a pod allows an attacker to compromise and control it """

    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Patched A Pod",
                               category=AccessRisk)
        self.evidence = evidence


class DeleteAPod(Vulnerability, Event):
    """ Deleting a pod allows an attacker to disturb applications on the cluster """

    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Deleted A Pod",
                               category=AccessRisk)
        self.evidence = evidence


class ApiServerPassiveHunterFinished(Event):
    def __init__(self, namespaces):
        self.namespaces = namespaces


# This Hunter checks what happens if we try to access the API Server without a service account token
# If we have a service account token we'll also trigger AccessApiServerWithToken below
@handler.subscribe(ApiServer)
class AccessApiServer(Hunter):
    """ API Server Hunter
    Checks if API server is accessible
    """

    def __init__(self, event):
        self.event = event
        self.path = "{}://{}:{}".format(self.event.protocol, self.event.host, self.event.port)
        self.headers = {}
        self.with_token = False

    def access_api_server(self):
        logging.debug('Passive Hunter is attempting to access the API at {}'.format(self.path))
        try:
            r = requests.get("{path}/api".format(path=self.path), headers=self.headers, verify=False)
            if r.status_code == 200 and r.content != '':
                return r.content
        except requests.exceptions.ConnectionError:
            pass
        return False

    def get_items(self, path):
        try:
            items = []
            r = requests.get(path, headers=self.headers, verify=False)
            if r.status_code ==200:
                resp = json.loads(r.content)
                for item in resp["items"]:
                    items.append(item["metadata"]["name"])
                return items
            logging.debug("Got HTTP {} respone: {}".format(r.status_code, r.text))
        except (requests.exceptions.ConnectionError, KeyError):
            logging.debug("Failed retrieving items from API server at {}".format(path))

        return None

    def get_pods(self, namespace=None):
        pods = []
        try:
            if namespace is None:
                r = requests.get("{path}/api/v1/pods".format(path=self.path),
                               headers=self.headers, verify=False)
            else:
                r = requests.get("{path}/api/v1/namespaces/{namespace}/pods".format(path=self.path),
                               headers=self.headers, verify=False)
            if r.status_code == 200:
                resp = json.loads(r.content)
                for item in resp["items"]:
                    name = item["metadata"]["name"].encode('ascii', 'ignore')
                    namespace = item["metadata"]["namespace"].encode('ascii', 'ignore')
                    pods.append({'name': name, 'namespace': namespace})

                return pods
        except (requests.exceptions.ConnectionError, KeyError):
            pass
        return None

    def execute(self):
        api = self.access_api_server()
        if api:
            if self.event.protocol == "http":
                self.publish_event(ServerApiHTTPAccess(api))
            else:
                self.publish_event(ServerApiAccess(api, self.with_token))

        namespaces = self.get_items("{path}/api/v1/namespaces".format(path=self.path))
        if namespaces:
            self.publish_event(ListNamespaces(namespaces, self.with_token))

        roles = self.get_items("{path}/apis/rbac.authorization.k8s.io/v1/roles".format(path=self.path))
        if roles:
            self.publish_event(ListRoles(roles, self.with_token))

        cluster_roles = self.get_items("{path}/apis/rbac.authorization.k8s.io/v1/clusterroles".format(path=self.path))
        if cluster_roles:
            self.publish_event(ListClusterRoles(cluster_roles, self.with_token))

        pods = self.get_pods()
        if pods:
            self.publish_event(ListPodsAndNamespaces(pods, self.with_token))

        # If we have a service account token, this event should get triggered twice - once with and once without
        # the token
        self.publish_event(ApiServerPassiveHunterFinished(namespaces))

@handler.subscribe(ApiServer, predicate=lambda x: x.auth_token)
class AccessApiServerWithToken(AccessApiServer):
    """ API Server Hunter
    Accessing the API server using the service account token obtained from a compromised pod
    """

    def __init__(self, event):
        super(AccessApiServerWithToken, self).__init__(event)
        assert self.event.auth_token != ''
        self.headers = {'Authorization': 'Bearer ' + self.event.auth_token}
        self.category = InformationDisclosure
        self.with_token = True


# Active Hunter
@handler.subscribe(ApiServerPassiveHunterFinished)
class AccessApiServerActive(ActiveHunter):
    """API server hunter
    Accessing the api server might grant an attacker full control over the cluster
    """

    def __init__(self, event):
        self.event = event
        self.path = "{}://{}:{}".format(self.event.protocol, self.event.host, self.event.port)

    def create_item(self, path, name, data):
        headers = {
            'Content-Type': 'application/json'
        }
        if self.event.auth_token:
            headers['Authorization'] = 'Bearer {token}'.format(token=self.event.auth_token)

        try:
            res = requests.post(path.format(name=name), verify=False, data=data, headers=headers)
            if res.status_code in [200, 201, 202]:
                parsed_content = json.loads(res.content)
                return parsed_content['metadata']['name']
        except (requests.exceptions.ConnectionError, KeyError):
            pass
        return None

    def patch_item(self, path, data):
        headers = {
            'Content-Type': 'application/json-patch+json'
        }
        if self.event.auth_token:
            headers['Authorization'] = 'Bearer {token}'.format(token=self.event.auth_token)
        try:
            res = requests.patch(path, headers=headers, verify=False, data=data)
            if res.status_code not in [200, 201, 202]:
                return None
            parsed_content = json.loads(res.content)
            # TODO is there a patch timestamp we could use?
            return parsed_content['metadata']['namespace']
        except (requests.exceptions.ConnectionError, KeyError):
            pass
        return None

    def delete_item(self, path):
        headers = {}
        if self.event.auth_token:
            headers['Authorization'] = 'Bearer {token}'.format(token=self.event.auth_token)
        try:
            res = requests.delete(path, headers=headers, verify=False)
            if res.status_code in [200, 201, 202]:
                parsed_content = json.loads(res.content)
                return parsed_content['metadata']['deletionTimestamp']
        except (requests.exceptions.ConnectionError, KeyError):
            pass
        return None

    def create_a_pod(self, namespace, is_privileged):
        privileged_value = ',"securityContext":{"privileged":true}' if is_privileged else ''
        random_name = (str(uuid.uuid4()))[0:5]
        json_pod = \
            """
                {{"apiVersion": "v1",
                "kind": "Pod",
                "metadata": {{
                    "name": "{random_name}"
                }},
                "spec": {{
                    "containers": [
                        {{
                            "name": "{random_name}",
                            "image": "nginx:1.7.9",
                            "ports": [
                                {{
                                    "containerPort": 80
                                }}
                            ]
                            {is_privileged_flag}
                        }}
                    ]
                }}
            }}
            """.format(random_name=random_name, is_privileged_flag=privileged_value)
        return self.create_item(path="{path}/api/v1/namespaces/{namespace}/pods".format(
                                path=self.path, namespace=namespace), name=random_name, data=json_pod)

    def delete_a_pod(self, namespace, pod_name):
        delete_timestamp = self.delete_item("{path}/api/v1/namespaces/{namespace}/pods/{name}".format(
                                          path=self.path, name=pod_name, namespace=namespace))
        if delete_timestamp is None:
            logging.error("Created pod {name} in namespace {namespace} but unable to delete it".format(name=pod_name, namespace=namespace))
        return delete_timestamp

    def patch_a_pod(self, namespace, pod_name):
        data = '[{ "op": "add", "path": "/hello", "value": ["world"] }]'
        return self.patch_item(path="{path}/api/v1/namespaces/{namespace}/pods/{name}".format(
                                 path=self.path, namespace=namespace, name=pod_name),
                                 data=data)

    def create_namespace(self):
        random_name = (str(uuid.uuid4()))[0:5]
        json = '{{"kind":"Namespace","apiVersion":"v1","metadata":{{"name":"{random_str}","labels":{{"name":"{random_str}"}}}}}}'.format(random_str=random_name)
        return self.create_item(path="{path}/api/v1/namespaces".format(path=self.path), name=random_name, data=json)

    def delete_namespace(self, namespace):
        delete_timestamp = self.delete_item("{path}/api/v1/namespaces/{name}".format(path=self.path, name=namespace))
        if delete_timestamp is None:
            logging.error("Created namespace {namespace} but unable to delete it".format(namespace=namespace))
        return delete_timestamp

    def create_a_role(self, namespace):
        name = (str(uuid.uuid4()))[0:5]
        role = """{{
                          "kind": "Role",
                          "apiVersion": "rbac.authorization.k8s.io/v1",
                          "metadata": {{
                            "namespace": "{namespace}",
                            "name": "{random_str}"
                          }},
                          "rules": [
                            {{
                              "apiGroups": [
                                ""
                              ],
                              "resources": [
                                "pods"
                              ],
                              "verbs": [
                                "get",
                                "watch",
                                "list"
                              ]
                            }}
                          ]
                        }}""".format(random_str=name, namespace=namespace)
        return self.create_item(path="{path}/apis/rbac.authorization.k8s.io/v1/namespaces/{namespace}/roles".format(
                                path=self.path, namespace=namespace), name=name, data=role)

    def create_a_cluster_role(self):
        name = (str(uuid.uuid4()))[0:5]
        cluster_role = """{{
                      "kind": "ClusterRole",
                      "apiVersion": "rbac.authorization.k8s.io/v1",
                      "metadata": {{
                        "name": "{random_str}"
                      }},
                      "rules": [
                        {{
                          "apiGroups": [
                            ""
                          ],
                          "resources": [
                            "pods"
                          ],
                          "verbs": [
                            "get",
                            "watch",
                            "list"
                          ]
                        }}
                      ]
                    }}""".format(random_str=name)
        return self.create_item(path="{path}/apis/rbac.authorization.k8s.io/v1/clusterroles".format(
                               path=self.path), name=name, data=cluster_role)

    def delete_a_role(self, namespace, name):
        delete_timestamp = self.delete_item("{path}/apis/rbac.authorization.k8s.io/v1/namespaces/{namespace}/roles/{role}".format(
            path=self.path, name=namespace, role=name))
        if delete_timestamp is None:
            logging.error("Created role {name} in namespace {namespace} but unable to delete it".format(name=name, namespace=namespace))
        return delete_timestamp

    def delete_a_cluster_role(self, name):
        delete_timestamp = self.delete_item("{path}/apis/rbac.authorization.k8s.io/v1/clusterroles/{role}".format(
            path=self.path, role=name))
        if delete_timestamp is None:
            logging.error("Created cluster role {name} but unable to delete it".format(name=name))
        return delete_timestamp

    def patch_a_role(self, namespace, role):
        data = '[{ "op": "add", "path": "/hello", "value": ["world"] }]'
        return self.patch_item(path="{path}/apis/rbac.authorization.k8s.io/v1/namespaces/{namespace}/roles/{name}".format(
                                path=self.path, name=role, namespace=namespace),
                                data=data)

    def patch_a_cluster_role(self, cluster_role):
        data = '[{ "op": "add", "path": "/hello", "value": ["world"] }]'
        return self.patch_item(path="{path}/apis/rbac.authorization.k8s.io/v1/clusterroles/{name}".format(
                                path=self.path, name=cluster_role),
                                data=data)

    def execute(self):
        # Try creating cluster-wide objects
        namespace = self.create_namespace()
        if namespace:
            self.publish_event(CreateANamespace('new namespace name: {name}'.format(name=namespace)))
            delete_timestamp = self.delete_namespace(namespace)
            if delete_timestamp:
                self.publish_event(DeleteANamespace(delete_timestamp))

        cluster_role = self.create_a_cluster_role()
        if cluster_role:
            self.publish_event(CreateAClusterRole('Cluster role name:  {name}'.format(name=cluster_role)))

            patch_evidence = self.patch_a_cluster_role(cluster_role)
            if patch_evidence:
                self.publish_event(PatchAClusterRole('Patched Cluster Role Name: {name}  Patch evidence: {patch_evidence}'.format(
                    name=cluster_role, patch_evidence=patch_evidence)))

            delete_timestamp = self.delete_a_cluster_role(cluster_role)
            if delete_timestamp:
                self.publish_event(DeleteAClusterRole('Cluster role {name} deletion time {delete_timestamp}'.format(
                                                        name=cluster_role, delete_timestamp=delete_timestamp)))

        #  Try attacking all the namespaces we know about
        if self.event.namespaces:
            for namespace in self.event.namespaces:
                # Try creating and deleting a privileged pod
                pod_name = self.create_a_pod(namespace, True)
                if pod_name:
                    self.publish_event(CreateAPrivilegedPod('Pod Name: {pod_name}  Namespace: {namespace}'.format(
                                                    pod_name=pod_name, namespace=namespace)))
                    delete_time = self.delete_a_pod(namespace, pod_name)
                    if delete_time:
                        self.publish_event(DeleteAPod('Pod Name: {pod_name}  deletion time: {delete_time}'.format(
                                                    pod_name=pod_name, delete_evidence=delete_time)))

                # Try creating, patching and deleting an unprivileged pod
                pod_name = self.create_a_pod(namespace, False)
                if pod_name:
                    self.publish_event(CreateAPod('Pod Name: {pod_name}  Namespace: {namespace}'.format(
                                                    pod_name=pod_name, namespace=namespace)))

                    patch_evidence = self.patch_a_pod(namespace, pod_name)
                    if patch_evidence:
                        self.publish_event(PatchAPod('Pod Name: {pod_name}  Namespace: {namespace}  Patch evidence: {patch_evidence}'.format(
                                                        pod_name=pod_name, namespace=namespace,
                                                        patch_evidence=patch_evidence)))

                    delete_time = self.delete_a_pod(namespace, pod_name)
                    if delete_time:
                        self.publish_event(DeleteAPod('Pod Name: {pod_name}  Namespace: {namespace}  Delete time: {delete_time}'.format(
                                                        pod_name=pod_name, namespace=namespace, delete_time=delete_time)))

                role = self.create_a_role(namespace)
                if role:
                    self.publish_event(CreateARole('Role name:  {name}'.format(name=role)))

                    patch_evidence = self.patch_a_role(namespace, role)
                    if patch_evidence:
                        self.publish_event(PatchARole('Patched Role Name: {name}  Namespace: {namespace}  Patch evidence: {patch_evidence}'.format(
                            name=role, namespace=namespace, patch_evidence=patch_evidence)))

                    delete_time = self.delete_a_role(namespace, role)
                    if delete_time:
                        self.publish_event(DeleteARole('Deleted role: {name}  Namespace: {namespace}  Delete time: {delete_time}'.format(
                                                        name=role, namespace=namespace, delete_time=delete_time)))


            #  Note: we are not binding any role or cluster role because
            # -- in certain cases it might effect the running pod within the cluster (and we don't want to do that).

@handler.subscribe(ApiServer)
class ApiVersionHunter(Hunter):
    """Api Version Hunter
    Tries to obtain the Api Server's version directly from /version endpoint
    """
    def __init__(self, event):
        self.event = event
        self.path = "{}://{}:{}".format(self.event.protocol, self.event.host, self.event.port)
        self.session = requests.Session()
        self.session.verify = False
        if self.event.auth_token:
            self.session.headers.update({"Authorization": "Bearer {}".format(self.event.auth_token)})

    def execute(self):
        if self.event.auth_token:
            logging.debug('Passive Hunter is attempting to access the API server version end point using the pod\'s service account token on {}:{} \t'.format(self.event.host, self.event.port))
        else:
            logging.debug('Passive Hunter is attempting to access the API server version end point anonymously')
        version = json.loads(self.session.get(self.path + "/version").text)["gitVersion"]
        logging.debug("Discovered version of api server {}".format(version))
        self.publish_event(K8sVersionDisclosure(version=version, from_endpoint="/version"))



from os.path import dirname, basename, isfile
import glob

# dynamically importing all modules in folder
files = glob.glob(dirname(__file__)+"/*.py")
for module_name in (basename(f)[:-3] for f in files if isfile(f) and not f.endswith('__init__.py')):
    if not module_name.startswith('test_'):
        exec('from .{} import *'.format(module_name))



from argparse import ArgumentParser


def parse_args():
    parser = ArgumentParser(
        description='Kube-Hunter - hunts for security '
                    'weaknesses in Kubernetes clusters')

    parser.add_argument(
        '--list',
        action="store_true",
        help="Displays all tests in kubehunter "
             "(add --active flag to see active tests)")

    parser.add_argument(
        '--interface',
        action="store_true",
        help="Set hunting on all network interfaces")

    parser.add_argument(
        '--pod',
        action="store_true",
        help="Set hunter as an insider pod")

    parser.add_argument(
        '--quick',
        action="store_true",
        help="Prefer quick scan (subnet 24)")

    parser.add_argument(
        '--include-patched-versions',
        action="store_true",
        help="Don't skip patched versions when scanning")

    parser.add_argument(
        '--cidr',
        type=str,
        help="Set an ip range to scan, example: 192.168.0.0/16")
    parser.add_argument(
        '--mapping',
        action="store_true",
        help="Outputs only a mapping of the cluster's nodes")

    parser.add_argument(
        '--remote',
        nargs='+',
        metavar="HOST",
        default=list(),
        help="One or more remote ip/dns to hunt")

    parser.add_argument(
        '--active',
        action="store_true",
        help="Enables active hunting")

    parser.add_argument(
        '--log',
        type=str,
        metavar="LOGLEVEL",
        default='INFO',
        help="Set log level, options are: debug, info, warn, none")

    parser.add_argument(
        '--report',
        type=str,
        default='plain',
        help="Set report type, options are: plain, yaml, json")

    parser.add_argument(
        '--dispatch',
        type=str,
        default='stdout',
        help="Where to send the report to, options are: "
        "stdout, http (set KUBEHUNTER_HTTP_DISPATCH_URL and "
        "KUBEHUNTER_HTTP_DISPATCH_METHOD environment variables to configure)")

    parser.add_argument(
        '--statistics',
        action="store_true",
        help="Show hunting statistics")

    return parser.parse_args()



import logging
import requests

from collections import defaultdict
from requests import get

from kube_hunter.core.types import Discovery
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Service, Event, OpenPortEvent

class KubeProxyEvent(Event, Service):
    """proxies from a localhost address to the Kubernetes apiserver"""
    def __init__(self):
        Service.__init__(self, name="Kubernetes Proxy")

@handler.subscribe(OpenPortEvent, predicate=lambda x: x.port == 8001)
class KubeProxy(Discovery):
    """Proxy Discovery
    Checks for the existence of a an open Proxy service
    """
    def __init__(self, event):
        self.event = event
        self.host = event.host
        self.port = event.port or 8001

    @property
    def accesible(self):
        logging.debug("Attempting to discover a proxy service")
        r = requests.get("http://{host}:{port}/api/v1".format(host=self.host, port=self.port))
        if r.status_code == 200 and "APIResourceList" in r.text:
            return True

    def execute(self):
        if self.accesible:
            self.publish_event(KubeProxyEvent())



import json
import logging

import requests

from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Event, OpenPortEvent, Service
from kube_hunter.core.types import Discovery


class KubeDashboardEvent(Service, Event):
    """A web-based Kubernetes user interface. allows easy usage with operations on the cluster"""
    def __init__(self, **kargs):
        Service.__init__(self, name="Kubernetes Dashboard", **kargs)

@handler.subscribe(OpenPortEvent, predicate=lambda x: x.port == 30000)
class KubeDashboard(Discovery):
    """K8s Dashboard Discovery
    Checks for the existence of a Dashboard
    """
    def __init__(self, event):
        self.event = event

    @property
    def secure(self):
        logging.debug("Attempting to discover an Api server to access dashboard")
        r = requests.get("http://{}:{}/api/v1/service/default".format(self.event.host, self.event.port))
        if "listMeta" in r.text and len(json.loads(r.text)["errors"]) == 0:
            return False
        return True

    def execute(self):
        if not self.secure:
            self.publish_event(KubeDashboardEvent())



import logging
import json
import requests

from kube_hunter.core.types import Hunter, RemoteCodeExec, KubernetesCluster
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Vulnerability, Event
from kube_hunter.modules.discovery.dashboard import KubeDashboardEvent


class DashboardExposed(Vulnerability, Event):
    """All operations on the cluster are exposed"""
    def __init__(self, nodes):
        Vulnerability.__init__(self, KubernetesCluster, "Dashboard Exposed", category=RemoteCodeExec, vid="KHV029")
        self.evidence = "nodes: {}".format(' '.join(nodes)) if nodes else None

@handler.subscribe(KubeDashboardEvent)
class KubeDashboard(Hunter):
    """Dashboard Hunting
    Hunts open Dashboards, gets the type of nodes in the cluster
    """
    def __init__(self, event):
        self.event = event

    def get_nodes(self):
        logging.debug("Passive hunter is attempting to get nodes types of the cluster")
        r = requests.get("http://{}:{}/api/v1/node".format(self.event.host, self.event.port))
        if r.status_code == 200 and "nodes" in r.text:
            return list(map(lambda node: node["objectMeta"]["name"], json.loads(r.text)["nodes"]))

    def execute(self):
        self.publish_event(DashboardExposed(nodes=self.get_nodes()))



import json
import requests
import logging

from kube_hunter.core.types import Discovery
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import OpenPortEvent, Service, Event, EventFilterBase

KNOWN_API_PORTS = [443, 6443, 8080]

class K8sApiService(Service, Event):
    """A Kubernetes API service"""
    def __init__(self, protocol="https"):
        Service.__init__(self, name="Unrecognized K8s API")
        self.protocol = protocol


class ApiServer(Service, Event):
    """The API server is in charge of all operations on the cluster."""
    def __init__(self):
        Service.__init__(self, name="API Server")
        self.protocol = "https"

class MetricsServer(Service, Event):
    """The Metrics server is in charge of providing resource usage metrics for pods and nodes to the API server."""
    def __init__(self):
        Service.__init__(self, name="Metrics Server")
        self.protocol = "https"


# Other devices could have this port open, but we can check to see if it looks like a Kubernetes api
# A Kubernetes API service will respond with a JSON message that includes a "code" field for the HTTP status code
@handler.subscribe(OpenPortEvent, predicate=lambda x: x.port in KNOWN_API_PORTS)
class ApiServiceDiscovery(Discovery):
    """API Service Discovery
    Checks for the existence of K8s API Services
    """
    def __init__(self, event):
        self.event = event
        self.session = requests.Session()
        self.session.verify = False

    def execute(self):
        logging.debug("Attempting to discover an API service on {}:{}".format(self.event.host, self.event.port))
        protocols = ["http", "https"]
        for protocol in protocols:
            if self.has_api_behaviour(protocol):
                self.publish_event(K8sApiService(protocol))

    def has_api_behaviour(self, protocol):
        try:
            r = self.session.get("{}://{}:{}".format(protocol, self.event.host, self.event.port))
            if ('k8s' in r.text) or ('"code"' in r.text and r.status_code != 200):
                return True
        except requests.exceptions.SSLError:
            logging.debug("{} protocol not accepted on {}:{}".format(protocol, self.event.host, self.event.port))
        except Exception as e:
            logging.debug("{} on {}:{}".format(e, self.event.host, self.event.port))


# Acts as a Filter for services, In the case that we can classify the API,
# We swap the filtered event with a new corresponding Service to next be published
# The classification can be regarding the context of the execution,
# Currently we classify: Metrics Server and Api Server
# If running as a pod:
# We know the Api server IP, so we can classify easily
# If not:
# We determine by accessing the /version on the service.
# Api Server will contain a major version field, while the Metrics will not
@handler.subscribe(K8sApiService)
class ApiServiceClassify(EventFilterBase):
    """API Service Classifier
    Classifies an API service
    """
    def __init__(self, event):
        self.event = event
        self.classified = False
        self.session = requests.Session()
        self.session.verify = False
        # Using the auth token if we can, for the case that authentication is needed for our checks
        if self.event.auth_token:
            self.session.headers.update({"Authorization": "Bearer {}".format(self.event.auth_token)})

    def classify_using_version_endpoint(self):
        """Tries to classify by accessing /version. if could not access succeded, returns"""
        try:
            r = self.session.get("{}://{}:{}/version".format(self.event.protocol, self.event.host, self.event.port))
            versions = r.json()
            if 'major' in versions:
                if versions.get('major') == "":
                    self.event = MetricsServer()
                else:
                    self.event = ApiServer()
        except Exception as e:
            logging.exception("Could not access /version on API service")

    def execute(self):
        discovered_protocol = self.event.protocol
        # if running as pod
        if self.event.kubeservicehost:
            # if the host is the api server's IP, we know it's the Api Server
            if self.event.kubeservicehost == str(self.event.host):
                self.event = ApiServer()
            else:
                self.event = MetricsServer()
        # if not running as pod.
        else:
            self.classify_using_version_endpoint()

        # in any case, making sure to link previously discovered protocol
        self.event.protocol = discovered_protocol
        # If some check classified the Service,
        # the event will have been replaced.
        return self.event



from kube_hunter.core.types import Discovery
from kube_hunter.modules.report.collector import services, vulnerabilities, \
    hunters, services_lock, vulnerabilities_lock


class BaseReporter(object):
    def get_nodes(self):
        nodes = list()
        node_locations = set()
        with services_lock:
            for service in services:
                node_location = str(service.host)
                if node_location not in node_locations:
                    nodes.append({
                        "type": "Node/Master",
                        "location": str(service.host)
                    })
                    node_locations.add(node_location)
        return nodes

    def get_services(self):
        with services_lock:
            services_data = [{
                "service": service.get_name(),
                "location": f"{service.host}:"
                            f"{service.port}"
                            f"{service.get_path()}",
                "description": service.explain()
            } for service in services]
        return services_data

    def get_vulnerabilities(self):
        with vulnerabilities_lock:
            vulnerabilities_data = [{
                "location": vuln.location(),
                "vid": vuln.get_vid(),
                "category": vuln.category.name,
                "severity": vuln.get_severity(),
                "vulnerability": vuln.get_name(),
                "description": vuln.explain(),
                "evidence": str(vuln.evidence),
                "hunter": vuln.hunter.get_name()
            } for vuln in vulnerabilities]
        return vulnerabilities_data

    def get_hunter_statistics(self):
        hunters_data = list()
        for hunter, docs in hunters.items():
            if not Discovery in hunter.__mro__:
                name, doc = hunter.parse_docs(docs)
                hunters_data.append({
                    "name": name,
                    "description": doc,
                    "vulnerabilities": hunter.publishedVulnerabilities
                })
        return hunters_data

    def get_report(self, *, statistics, **kwargs):
        report = {
            "nodes": self.get_nodes(),
            "services": self.get_services(),
            "vulnerabilities": self.get_vulnerabilities()
        }

        if statistics:
            report["hunter_statistics"] = self.get_hunter_statistics()

        report["kburl"] = "https://aquasecurity.github.io/kube-hunter/kb/{vid}"

        return report



import requests_mock
import time
from queue import Empty

from kube_hunter.modules.discovery.hosts import FromPodHostDiscovery, RunningAsPodEvent, HostScanEvent, AzureMetadataApi
from kube_hunter.core.events.types import Event, NewHostEvent
from kube_hunter.core.events import handler
from kube_hunter.conf import config

def test_FromPodHostDiscovery():

    with requests_mock.Mocker() as m:
        e = RunningAsPodEvent()

        config.azure = False
        config.remote = None
        config.cidr = None
        m.get("http://169.254.169.254/metadata/instance?api-version=2017-08-01", status_code=404)
        f = FromPodHostDiscovery(e)
        assert not f.is_azure_pod()
        # TODO For now we don't test the traceroute discovery version
        # f.execute()

        # Test that we generate NewHostEvent for the addresses reported by the Azure Metadata API
        config.azure = True
        m.get("http://169.254.169.254/metadata/instance?api-version=2017-08-01", \
            text='{"network":{"interface":[{"ipv4":{"subnet":[{"address": "3.4.5.6", "prefix": "255.255.255.252"}]}}]}}')
        assert f.is_azure_pod()
        f.execute()

        # Test that we don't trigger a HostScanEvent unless either config.remote or config.cidr are configured
        config.remote = "1.2.3.4"
        f.execute()

        config.azure = False
        config.remote = None 
        config.cidr = "1.2.3.4/24"
        f.execute()


# In this set of tests we should only trigger HostScanEvent when remote or cidr are set
@handler.subscribe(HostScanEvent)
class testHostDiscovery(object):
    def __init__(self, event):
        assert config.remote is not None or config.cidr is not None
        assert config.remote == "1.2.3.4" or config.cidr == "1.2.3.4/24"
        
# In this set of tests we should only get as far as finding a host if it's Azure
# because we're not running the code that would normally be triggered by a HostScanEvent
@handler.subscribe(NewHostEvent)
class testHostDiscoveryEvent(object):
    def __init__(self, event):
        assert config.azure
        assert str(event.host).startswith("3.4.5.")
        assert config.remote is None 
        assert config.cidr is None

# Test that we only report this event for Azure hosts
@handler.subscribe(AzureMetadataApi)
class testAzureMetadataApi(object):
    def __init__(self, event):
        assert config.azure



import ssl
import logging
import base64
import re

from socket import socket

from kube_hunter.core.types import Hunter, KubernetesCluster, InformationDisclosure
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Vulnerability, Event, Service


email_pattern = re.compile(r"([a-z0-9]+@[a-z0-9]+\.[a-z0-9]+)")

class CertificateEmail(Vulnerability, Event):
    """Certificate includes an email address"""
    def __init__(self, email):
        Vulnerability.__init__(self, KubernetesCluster, "Certificate Includes Email Address", category=InformationDisclosure,khv="KHV021")
        self.email = email
        self.evidence = "email: {}".format(self.email)

@handler.subscribe(Service)
class CertificateDiscovery(Hunter):
    """Certificate Email Hunting
    Checks for email addresses in kubernetes ssl certificates
    """
    def __init__(self, event):
        self.event = event

    def execute(self):
        try:
            logging.debug("Passive hunter is attempting to get server certificate")
            addr = (str(self.event.host), self.event.port)
            cert = ssl.get_server_certificate(addr)
        except ssl.SSLError:
            # If the server doesn't offer SSL on this port we won't get a certificate
            return
        c = cert.strip(ssl.PEM_HEADER).strip(ssl.PEM_FOOTER)
        certdata = base64.decodestring(c)
        emails = re.findall(email_pattern, certdata)
        for email in emails:
            self.publish_event( CertificateEmail(email=email) )



import logging
import requests
import json

from enum import Enum

from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Event, Vulnerability, K8sVersionDisclosure
from kube_hunter.core.types import ActiveHunter, Hunter, KubernetesCluster, InformationDisclosure
from kube_hunter.modules.discovery.dashboard import KubeDashboardEvent
from kube_hunter.modules.discovery.proxy import KubeProxyEvent


class KubeProxyExposed(Vulnerability, Event):
    """All operations on the cluster are exposed"""
    def __init__(self):
        Vulnerability.__init__(self, KubernetesCluster, "Proxy Exposed", category=InformationDisclosure, vid="KHV049")

class Service(Enum):
    DASHBOARD = "kubernetes-dashboard"

@handler.subscribe(KubeProxyEvent)
class KubeProxy(Hunter):
    """Proxy Hunting
    Hunts for a dashboard behind the proxy
    """
    def __init__(self, event):
        self.event = event
        self.api_url = "http://{host}:{port}/api/v1".format(host=self.event.host, port=self.event.port)

    def execute(self):
        self.publish_event(KubeProxyExposed())
        for namespace, services in self.services.items():
            for service in services:
                if service == Service.DASHBOARD.value:
                    logging.debug(service)
                    curr_path = "api/v1/namespaces/{ns}/services/{sv}/proxy".format(ns=namespace,sv=service) # TODO: check if /proxy is a convention on other services
                    self.publish_event(KubeDashboardEvent(path=curr_path, secure=False))

    @property
    def namespaces(self):
        resource_json = requests.get(self.api_url + "/namespaces").json()
        return self.extract_names(resource_json)

    @property
    def services(self):
        # map between namespaces and service names
        services = dict()
        for namespace in self.namespaces:
            resource_path = "/namespaces/{ns}/services".format(ns=namespace)
            resource_json = requests.get(self.api_url + resource_path).json()
            services[namespace] = self.extract_names(resource_json)
        logging.debug(services)
        return services

    @staticmethod
    def extract_names(resource_json):
        names = list()
        for item in resource_json["items"]:
            names.append(item["metadata"]["name"])
        return names

@handler.subscribe(KubeProxyExposed)
class ProveProxyExposed(ActiveHunter):
    """Build Date Hunter
    Hunts when proxy is exposed, extracts the build date of kubernetes
    """
    def __init__(self, event):
        self.event = event

    def execute(self):
        version_metadata = json.loads(requests.get("http://{host}:{port}/version".format(
            host=self.event.host,
            port=self.event.port,
        ), verify=False).text)
        if "buildDate" in version_metadata:
            self.event.evidence = "build date: {}".format(version_metadata["buildDate"])

@handler.subscribe(KubeProxyExposed)
class K8sVersionDisclosureProve(ActiveHunter):
    """K8s Version Hunter
    Hunts Proxy when exposed, extracts the version
    """
    def __init__(self, event):
        self.event = event

    def execute(self):
        version_metadata = json.loads(requests.get("http://{host}:{port}/version".format(
            host=self.event.host,
            port=self.event.port,
        ), verify=False).text)
        if "gitVersion" in version_metadata:
            self.publish_event(K8sVersionDisclosure(version=version_metadata["gitVersion"], from_endpoint="/version", extra_info="on the kube-proxy"))



from __future__ import print_function

from prettytable import ALL, PrettyTable

from kube_hunter.modules.report.base import BaseReporter
from kube_hunter.modules.report.collector import services, vulnerabilities, \
    hunters, services_lock, vulnerabilities_lock

EVIDENCE_PREVIEW = 40
MAX_TABLE_WIDTH = 20
KB_LINK = "https://github.com/aquasecurity/kube-hunter/tree/master/docs/_kb"


class PlainReporter(BaseReporter):

    def get_report(self, *, statistics=None, mapping=None, **kwargs):
        """generates report tables"""
        output = ""

        with vulnerabilities_lock:
            vulnerabilities_len = len(vulnerabilities)

        hunters_len = len(hunters.items())

        with services_lock:
            services_len = len(services)

        if services_len:
            output += self.nodes_table()
            if not mapping:
                output += self.services_table()
                if vulnerabilities_len:
                    output += self.vulns_table()
                else:
                    output += "\nNo vulnerabilities were found"
                if statistics:
                    if hunters_len:
                        output += self.hunters_table()
                    else:
                        output += "\nNo hunters were found"
        else:
            if vulnerabilities_len:
                output += self.vulns_table()
            output += "\nKube Hunter couldn't find any clusters"
            # print("\nKube Hunter couldn't find any clusters. {}".format("Maybe try with --active?" if not config.active else ""))
        return output

    def nodes_table(self):
        nodes_table = PrettyTable(["Type", "Location"], hrules=ALL)
        nodes_table.align = "l"
        nodes_table.max_width = MAX_TABLE_WIDTH
        nodes_table.padding_width = 1
        nodes_table.sortby = "Type"
        nodes_table.reversesort = True
        nodes_table.header_style = "upper"
        id_memory = set()
        services_lock.acquire()
        for service in services:
            if service.event_id not in id_memory:
                nodes_table.add_row(["Node/Master", service.host])
                id_memory.add(service.event_id)
        nodes_ret = "\nNodes\n{}\n".format(nodes_table)
        services_lock.release()
        return nodes_ret

    def services_table(self):
        services_table = PrettyTable(["Service", "Location", "Description"], hrules=ALL)
        services_table.align = "l"
        services_table.max_width = MAX_TABLE_WIDTH
        services_table.padding_width = 1
        services_table.sortby = "Service"
        services_table.reversesort = True
        services_table.header_style = "upper"
        with services_lock:
            for service in services:
                services_table.add_row(
                    [service.get_name(),
                     f"{service.host}:"
                     f"{service.port}"
                     f"{service.get_path()}",
                     service.explain()])
            detected_services_ret = "\nDetected Services\n" \
                                    f"{services_table}\n"
        return detected_services_ret

    def vulns_table(self):
        column_names = ["ID", "Location",
                        "Category", "Vulnerability",
                        "Description", "Evidence"]
        vuln_table = PrettyTable(column_names, hrules=ALL)
        vuln_table.align = "l"
        vuln_table.max_width = MAX_TABLE_WIDTH
        vuln_table.sortby = "Category"
        vuln_table.reversesort = True
        vuln_table.padding_width = 1
        vuln_table.header_style = "upper"

        with vulnerabilities_lock:
            for vuln in vulnerabilities:
                evidence = str(vuln.evidence)[:EVIDENCE_PREVIEW] + "..." if len(str(vuln.evidence)) > EVIDENCE_PREVIEW else str(vuln.evidence)
                row = [vuln.get_vid(), vuln.location(), vuln.category.name, vuln.get_name(), vuln.explain(), evidence]
                vuln_table.add_row(row)
        return "\nVulnerabilities\nFor further information about a vulnerability, search its ID in: \n{}\n{}\n".format(KB_LINK, vuln_table)

    def hunters_table(self):
        column_names = ["Name", "Description", "Vulnerabilities"]
        hunters_table = PrettyTable(column_names, hrules=ALL)
        hunters_table.align = "l"
        hunters_table.max_width = MAX_TABLE_WIDTH
        hunters_table.sortby = "Name"
        hunters_table.reversesort = True
        hunters_table.padding_width = 1
        hunters_table.header_style = "upper"

        hunter_statistics = self.get_hunter_statistics()
        for item in hunter_statistics:
            hunters_table.add_row([item.get("name"), item.get("description"), item.get("vulnerabilities")])
        return "\nHunter Statistics\n{}\n".format(hunters_table)



import json
import logging
import requests
import urllib3

from enum import Enum

from kube_hunter.core.types import Discovery, Kubelet
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import OpenPortEvent, Vulnerability, Event, Service

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


""" Services """
class ReadOnlyKubeletEvent(Service, Event):
    """The read-only port on the kubelet serves health probing endpoints, and is relied upon by many kubernetes components"""
    def __init__(self):
        Service.__init__(self, name="Kubelet API (readonly)")

class SecureKubeletEvent(Service, Event):
    """The Kubelet is the main component in every Node, all pod operations goes through the kubelet"""
    def __init__(self, cert=False, token=False, anonymous_auth=True, **kwargs):
        self.cert = cert
        self.token = token
        self.anonymous_auth = anonymous_auth
        Service.__init__(self, name="Kubelet API", **kwargs)


class KubeletPorts(Enum):
    SECURED = 10250
    READ_ONLY = 10255

@handler.subscribe(OpenPortEvent, predicate= lambda x: x.port == 10255 or x.port == 10250)
class KubeletDiscovery(Discovery):
    """Kubelet Discovery
    Checks for the existence of a Kubelet service, and its open ports
    """
    def __init__(self, event):
        self.event = event

    def get_read_only_access(self):
        logging.debug("Passive hunter is attempting to get kubelet read access at {}:{}".format(self.event.host, self.event.port))
        r = requests.get("http://{host}:{port}/pods".format(host=self.event.host, port=self.event.port))
        if r.status_code == 200:
            self.publish_event(ReadOnlyKubeletEvent())

    def get_secure_access(self):
        logging.debug("Attempting to get kubelet secure access")
        ping_status = self.ping_kubelet()
        if ping_status == 200:
            self.publish_event(SecureKubeletEvent(secure=False))
        elif ping_status == 403:
            self.publish_event(SecureKubeletEvent(secure=True))
        elif ping_status == 401:
            self.publish_event(SecureKubeletEvent(secure=True, anonymous_auth=False))

    def ping_kubelet(self):
        logging.debug("Attempting to get pod info from kubelet")
        try:
            return requests.get("https://{host}:{port}/pods".format(host=self.event.host, port=self.event.port), verify=False).status_code
        except Exception as ex:
            logging.debug("Failed pinging https port 10250 on {} : {}".format(self.event.host, ex))

    def execute(self):
        if self.event.port == KubeletPorts.SECURED.value:
            self.get_secure_access()
        elif self.event.port == KubeletPorts.READ_ONLY.value:
            self.get_read_only_access()



import os
import json
import logging
import socket
import sys
import time
import requests

from enum import Enum
from netaddr import IPNetwork, IPAddress
from netifaces import AF_INET, ifaddresses, interfaces

from kube_hunter.conf import config
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Event, NewHostEvent, Vulnerability
from kube_hunter.core.types import Discovery, InformationDisclosure, Azure

class RunningAsPodEvent(Event):
    def __init__(self):
        self.name = 'Running from within a pod'
        self.auth_token = self.get_service_account_file("token")
        self.client_cert = self.get_service_account_file("ca.crt")
        self.namespace = self.get_service_account_file("namespace")
        self.kubeservicehost = os.environ.get("KUBERNETES_SERVICE_HOST", None)

    # Event's logical location to be used mainly for reports.
    def location(self):
        location = "Local to Pod"
        if 'HOSTNAME' in os.environ:
            location += "(" + os.environ['HOSTNAME'] + ")"

        return location

    def get_service_account_file(self, file):
        try:
            with open("/var/run/secrets/kubernetes.io/serviceaccount/{file}".format(file=file)) as f:
                return f.read()
        except IOError:
            pass

class AzureMetadataApi(Vulnerability, Event):
    """Access to the Azure Metadata API exposes information about the machines associated with the cluster"""
    def __init__(self, cidr):
        Vulnerability.__init__(self, Azure, "Azure Metadata Exposure", category=InformationDisclosure, vid="KHV003")
        self.cidr = cidr
        self.evidence = "cidr: {}".format(cidr)

class HostScanEvent(Event):
    def __init__(self, pod=False, active=False, predefined_hosts=list()):
        self.active = active # flag to specify whether to get actual data from vulnerabilities
        self.predefined_hosts = predefined_hosts

class HostDiscoveryHelpers:
    @staticmethod
    def get_cloud(host):
        try:
            logging.debug("Checking whether the cluster is deployed on azure's cloud")
            # azurespeed.com provide their API via HTTP only; the service can be queried with
            # HTTPS, but doesn't show a proper certificate. Since no encryption is worse then
            # any encryption, we go with the verify=false option for the time being. At least
            # this prevents leaking internal IP addresses to passive eavesdropping.
            # TODO: find a more secure service to detect cloud IPs
            metadata = requests.get("https://www.azurespeed.com/api/region?ipOrUrl={ip}".format(ip=host), verify=False).text
        except requests.ConnectionError as e:
            logging.info("- unable to check cloud: {0}".format(e))
            return
        if "cloud" in metadata:
            return json.loads(metadata)["cloud"]

    # generator, generating a subnet by given a cidr
    @staticmethod
    def generate_subnet(ip, sn="24"):
        logging.debug("HostDiscoveryHelpers.generate_subnet {0}/{1}".format(ip, sn))
        subnet = IPNetwork('{ip}/{sn}'.format(ip=ip, sn=sn))
        for ip in IPNetwork(subnet):
            logging.debug("HostDiscoveryHelpers.generate_subnet yielding {0}".format(ip))
            yield ip


@handler.subscribe(RunningAsPodEvent)
class FromPodHostDiscovery(Discovery):
    """Host Discovery when running as pod
    Generates ip adresses to scan, based on cluster/scan type
    """
    def __init__(self, event):
        self.event = event

    def execute(self):
        # Scan any hosts that the user specified
        if config.remote or config.cidr:
            self.publish_event(HostScanEvent())
        else:
            # Discover cluster subnets, we'll scan all these hosts
            if self.is_azure_pod():
                subnets, cloud = self.azure_metadata_discovery()
            else:
                subnets, cloud = self.traceroute_discovery()

            should_scan_apiserver = False
            if self.event.kubeservicehost:
                should_scan_apiserver = True
            for subnet in subnets:
                if self.event.kubeservicehost and self.event.kubeservicehost in IPNetwork("{}/{}".format(subnet[0], subnet[1])):
                    should_scan_apiserver = False
                logging.debug("From pod scanning subnet {0}/{1}".format(subnet[0], subnet[1]))
                for ip in HostDiscoveryHelpers.generate_subnet(ip=subnet[0], sn=subnet[1]):
                    self.publish_event(NewHostEvent(host=ip, cloud=cloud))
            if should_scan_apiserver:
                self.publish_event(NewHostEvent(host=IPAddress(self.event.kubeservicehost), cloud=cloud))

    def is_azure_pod(self):
        try:
            logging.debug("From pod attempting to access Azure Metadata API")
            if requests.get("http://169.254.169.254/metadata/instance?api-version=2017-08-01", headers={"Metadata":"true"}, timeout=5).status_code == 200:
                return True
        except requests.exceptions.ConnectionError:
            return False

   # for pod scanning
    def traceroute_discovery(self):
        external_ip = requests.get("http://canhazip.com").text # getting external ip, to determine if cloud cluster
        from scapy.all import ICMP, IP, Ether, srp1

        node_internal_ip = srp1(Ether() / IP(dst="google.com" , ttl=1) / ICMP(), verbose=0)[IP].src
        return [ [node_internal_ip,"24"], ], external_ip

    # querying azure's interface metadata api | works only from a pod
    def azure_metadata_discovery(self):
        logging.debug("From pod attempting to access azure's metadata")
        machine_metadata = json.loads(requests.get("http://169.254.169.254/metadata/instance?api-version=2017-08-01", headers={"Metadata":"true"}).text)
        address, subnet = "", ""
        subnets = list()
        for interface in machine_metadata["network"]["interface"]:
            address, subnet = interface["ipv4"]["subnet"][0]["address"], interface["ipv4"]["subnet"][0]["prefix"]
            logging.debug("From pod discovered subnet {0}/{1}".format(address, subnet if not config.quick else "24"))
            subnets.append([address,subnet if not config.quick else "24"])

            self.publish_event(AzureMetadataApi(cidr="{}/{}".format(address, subnet)))

        return subnets, "Azure"

@handler.subscribe(HostScanEvent)
class HostDiscovery(Discovery):
    """Host Discovery
    Generates ip adresses to scan, based on cluster/scan type
    """
    def __init__(self, event):
        self.event = event

    def execute(self):
        if config.cidr:
            try:
                ip, sn = config.cidr.split('/')
            except ValueError as e:
                logging.exception("unable to parse cidr")
                return
            cloud = HostDiscoveryHelpers.get_cloud(ip)
            for ip in HostDiscoveryHelpers.generate_subnet(ip, sn=sn):
                self.publish_event(NewHostEvent(host=ip, cloud=cloud))
        elif config.interface:
            self.scan_interfaces()
        elif len(config.remote) > 0:
            for host in config.remote:
                self.publish_event(NewHostEvent(host=host, cloud=HostDiscoveryHelpers.get_cloud(host)))

    # for normal scanning
    def scan_interfaces(self):
        try:
            logging.debug("HostDiscovery hunter attempting to get external IP address")
            external_ip = requests.get("http://canhazip.com").text # getting external ip, to determine if cloud cluster
        except requests.ConnectionError as e:
            logging.debug("unable to determine local IP address: {0}".format(e))
            logging.info("~ default to 127.0.0.1")
            external_ip = "127.0.0.1"
        cloud = HostDiscoveryHelpers.get_cloud(external_ip)
        for ip in self.generate_interfaces_subnet():
            handler.publish_event(NewHostEvent(host=ip, cloud=cloud))

    # generate all subnets from all internal network interfaces
    def generate_interfaces_subnet(self, sn='24'):
        for ifaceName in interfaces():
            for ip in [i['addr'] for i in ifaddresses(ifaceName).setdefault(AF_INET, [])]:
                if not self.event.localhost and InterfaceTypes.LOCALHOST.value in ip.__str__():
                    continue
                for ip in HostDiscoveryHelpers.generate_subnet(ip, sn):
                    yield ip

# for comparing prefixes
class InterfaceTypes(Enum):
    LOCALHOST = "127"



import logging
from kube_hunter.conf.parser import parse_args

config = parse_args()

loglevel = getattr(logging, config.log.upper(), None)

if not loglevel:
    logging.basicConfig(level=logging.INFO,
                        format='%(message)s',
                        datefmt='%H:%M:%S')
    logging.warning('Unknown log level selected, using info')
elif config.log.lower() != "none":
    logging.basicConfig(level=loglevel,
                        format='%(message)s',
                        datefmt='%H:%M:%S')

import plugins



import socket
import logging

from kube_hunter.modules.discovery.hosts import RunningAsPodEvent
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Event, Vulnerability
from kube_hunter.core.types import Hunter, AccessRisk, KubernetesCluster


class CapNetRawEnabled(Event, Vulnerability):
    """CAP_NET_RAW is enabled by default for pods. If an attacker manages to compromise a pod, they could potentially take advantage of this capability to perform network attacks on other pods running on the same node"""
    def __init__(self):
        Vulnerability.__init__(self, KubernetesCluster, name="CAP_NET_RAW Enabled", category=AccessRisk)


@handler.subscribe(RunningAsPodEvent)
class PodCapabilitiesHunter(Hunter):
    """Pod Capabilities Hunter
    Checks for default enabled capabilities in a pod
    """
    def __init__(self, event):
        self.event = event

    def check_net_raw(self):
        logging.debug("Passive hunter's trying to open a RAW socket")
        try:
            # trying to open a raw socket without CAP_NET_RAW will raise PermissionsError
            s = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_RAW)
            s.close()
            logging.debug("Passive hunter's closing RAW socket")
            return True
        except PermissionError:
            logging.debug("CAP_NET_RAW not enabled")

    def execute(self):
        if self.check_net_raw():
            self.publish_event(CapNetRawEnabled())



#!/usr/bin/env python3

import logging
import threading

from kube_hunter.conf import config
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import HuntFinished, HuntStarted
from kube_hunter.modules.discovery.hosts import RunningAsPodEvent, HostScanEvent
from kube_hunter.modules.report import get_reporter, get_dispatcher

config.reporter = get_reporter(config.report)
config.dispatcher = get_dispatcher(config.dispatch)

import kube_hunter


def interactive_set_config():
    """Sets config manually, returns True for success"""
    options = [("Remote scanning",
                "scans one or more specific IPs or DNS names"),
               ("Interface scanning",
                "scans subnets on all local network interfaces"),
               ("IP range scanning", "scans a given IP range")]

    print("Choose one of the options below:")
    for i, (option, explanation) in enumerate(options):
        print("{}. {} ({})".format(i+1, option.ljust(20), explanation))
    choice = input("Your choice: ")
    if choice == '1':
        config.remote = input("Remotes (separated by a ','): ").\
            replace(' ', '').split(',')
    elif choice == '2':
        config.interface = True
    elif choice == '3':
        config.cidr = input("CIDR (example - 192.168.1.0/24): ").\
            replace(' ', '')
    else:
        return False
    return True


def list_hunters():
    print("\nPassive Hunters:\n----------------")
    for hunter, docs in handler.passive_hunters.items():
        name, doc = hunter.parse_docs(docs)
        print("* {}\n  {}\n".format(name, doc))

    if config.active:
        print("\n\nActive Hunters:\n---------------")
        for hunter, docs in handler.active_hunters.items():
            name, doc = hunter.parse_docs(docs)
            print("* {}\n  {}\n".format( name, doc))


global hunt_started_lock
hunt_started_lock = threading.Lock()
hunt_started = False


def main():
    global hunt_started
    scan_options = [
        config.pod,
        config.cidr,
        config.remote,
        config.interface
    ]
    try:
        if config.list:
            list_hunters()
            return

        if not any(scan_options):
            if not interactive_set_config(): return

        with hunt_started_lock:
            hunt_started = True
        handler.publish_event(HuntStarted())
        if config.pod:
            handler.publish_event(RunningAsPodEvent())
        else:
            handler.publish_event(HostScanEvent())

        # Blocking to see discovery output
        handler.join()
    except KeyboardInterrupt:
        logging.debug("Kube-Hunter stopped by user")
    # happens when running a container without interactive option
    except EOFError:
        logging.error("\033[0;31mPlease run again with -it\033[0m")
    finally:
        hunt_started_lock.acquire()
        if hunt_started:
            hunt_started_lock.release()
            handler.publish_event(HuntFinished())
            handler.join()
            handler.free()
            logging.debug("Cleaned Queue")
        else:
            hunt_started_lock.release()


if __name__ == '__main__':
    main()



import logging
import json
import requests

from packaging import version

from kube_hunter.conf import config
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Vulnerability, Event, K8sVersionDisclosure
from kube_hunter.core.types import Hunter, ActiveHunter, KubernetesCluster, \
    RemoteCodeExec, AccessRisk, InformationDisclosure, PrivilegeEscalation, \
    DenialOfService, KubectlClient
from kube_hunter.modules.discovery.kubectl import KubectlClientEvent


""" Cluster CVES """
class ServerApiVersionEndPointAccessPE(Vulnerability, Event):
    """Node is vulnerable to critical CVE-2018-1002105"""
    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Critical Privilege Escalation CVE", category=PrivilegeEscalation, vid="KHV022")
        self.evidence = evidence

class ServerApiVersionEndPointAccessDos(Vulnerability, Event):
    """Node not patched for CVE-2019-1002100. Depending on your RBAC settings, a crafted json-patch could cause a Denial of Service."""
    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Denial of Service to Kubernetes API Server", category=DenialOfService, vid="KHV023")
        self.evidence = evidence

class PingFloodHttp2Implementation(Vulnerability, Event):
    """Node not patched for CVE-2019-9512. an attacker could cause a Denial of Service by sending specially crafted HTTP requests."""
    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Possible Ping Flood Attack", category=DenialOfService, vid="KHV024")
        self.evidence = evidence

class ResetFloodHttp2Implementation(Vulnerability, Event):
    """Node not patched for CVE-2019-9514. an attacker could cause a Denial of Service by sending specially crafted HTTP requests."""
    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Possible Reset Flood Attack", category=DenialOfService, vid="KHV025")
        self.evidence = evidence

class ServerApiClusterScopedResourcesAccess(Vulnerability, Event):
    """Api Server not patched for CVE-2019-11247. API server allows access to custom resources via wrong scope"""
    def __init__(self, evidence):
        Vulnerability.__init__(self, KubernetesCluster, name="Arbitrary Access To Cluster Scoped Resources", category=PrivilegeEscalation, vid="KHV026")
        self.evidence = evidence


""" Kubectl CVES """

class IncompleteFixToKubectlCpVulnerability(Vulnerability, Event):
    """The kubectl client is vulnerable to CVE-2019-11246, an attacker could potentially execute arbitrary code on the client's machine"""
    def __init__(self, binary_version):
        Vulnerability.__init__(self, KubectlClient, "Kubectl Vulnerable To CVE-2019-11246", category=RemoteCodeExec, vid="KHV027")
        self.binary_version = binary_version
        self.evidence = "kubectl version: {}".format(self.binary_version)

class KubectlCpVulnerability(Vulnerability, Event):
    """The kubectl client is vulnerable to CVE-2019-1002101, an attacker could potentially execute arbitrary code on the client's machine"""
    def __init__(self, binary_version):
        Vulnerability.__init__(self, KubectlClient, "Kubectl Vulnerable To CVE-2019-1002101", category=RemoteCodeExec, vid="KHV028")
        self.binary_version = binary_version
        self.evidence = "kubectl version: {}".format(self.binary_version)


class CveUtils:
    @staticmethod
    def get_base_release(full_ver):
        # if LegacyVersion, converting manually to a base version
        if type(full_ver) == version.LegacyVersion:
            return version.parse('.'.join(full_ver._version.split('.')[:2]))
        return version.parse('.'.join(map(str, full_ver._version.release[:2])))

    @staticmethod
    def to_legacy(full_ver):
        # converting version to version.LegacyVersion
        return version.LegacyVersion('.'.join(map(str, full_ver._version.release)))

    @staticmethod
    def to_raw_version(v):
        if type(v) != version.LegacyVersion:
            return '.'.join(map(str, v._version.release))
        return v._version

    @staticmethod
    def version_compare(v1, v2):
        """Function compares two versions, handling differences with conversion to LegacyVersion"""
        # getting raw version, while striping 'v' char at the start. if exists.
        # removing this char lets us safely compare the two version.
        v1_raw, v2_raw = CveUtils.to_raw_version(v1).strip('v'), CveUtils.to_raw_version(v2).strip('v')
        new_v1 = version.LegacyVersion(v1_raw)
        new_v2 = version.LegacyVersion(v2_raw)

        return CveUtils.basic_compare(new_v1, new_v2)

    @staticmethod
    def basic_compare(v1, v2):
        return (v1>v2)-(v1<v2)

    @staticmethod
    def is_downstream_version(version):
        return any(c in version for c in '+-~')

    @staticmethod
    def is_vulnerable(fix_versions, check_version, ignore_downstream=False):
        """Function determines if a version is vulnerable, by comparing to given fix versions by base release"""
        if ignore_downstream and CveUtils.is_downstream_version(check_version):
            return False

        vulnerable = False
        check_v = version.parse(check_version)
        base_check_v = CveUtils.get_base_release(check_v)

        # default to classic compare, unless the check_version is legacy.
        version_compare_func = CveUtils.basic_compare
        if type(check_v) == version.LegacyVersion:
            version_compare_func = CveUtils.version_compare

        if check_version not in fix_versions:
            # comparing ease base release for a fix
            for fix_v in fix_versions:
                fix_v = version.parse(fix_v)
                base_fix_v = CveUtils.get_base_release(fix_v)

                # if the check version and the current fix has the same base release
                if base_check_v == base_fix_v:
                    # when check_version is legacy, we use a custom compare func, to handle differences between versions.
                    if version_compare_func(check_v, fix_v) == -1:
                        # determine vulnerable if smaller and with same base version
                        vulnerable = True
                        break

        # if we did't find a fix in the fix releases, checking if the version is smaller that the first fix
        if not vulnerable and version_compare_func(check_v, version.parse(fix_versions[0])) == -1:
            vulnerable = True

        return vulnerable


@handler.subscribe_once(K8sVersionDisclosure)
class K8sClusterCveHunter(Hunter):
    """K8s CVE Hunter
    Checks if Node is running a Kubernetes version vulnerable to specific important CVEs
    """

    def __init__(self, event):
        self.event = event

    def execute(self):
        logging.debug('Api Cve Hunter determining vulnerable version: {}'.format(self.event.version))
        cve_mapping = {
            ServerApiVersionEndPointAccessPE: ["1.10.11", "1.11.5", "1.12.3"],
            ServerApiVersionEndPointAccessDos: ["1.11.8", "1.12.6", "1.13.4"],
            ResetFloodHttp2Implementation: ["1.13.10", "1.14.6", "1.15.3"],
            PingFloodHttp2Implementation: ["1.13.10", "1.14.6", "1.15.3"],
            ServerApiClusterScopedResourcesAccess: ["1.13.9", "1.14.5", "1.15.2"]
        }
        for vulnerability, fix_versions in cve_mapping.items():
            if CveUtils.is_vulnerable(fix_versions, self.event.version, not config.include_patched_versions):
                self.publish_event(vulnerability(self.event.version))


@handler.subscribe(KubectlClientEvent)
class KubectlCVEHunter(Hunter):
    """Kubectl CVE Hunter
    Checks if the kubectl client is vulnerable to specific important CVEs
    """
    def __init__(self, event):
        self.event = event

    def execute(self):
        cve_mapping = {
            KubectlCpVulnerability: ['1.11.9', '1.12.7', '1.13.5' '1.14.0'],
            IncompleteFixToKubectlCpVulnerability: ['1.12.9', '1.13.6', '1.14.2']
        }
        logging.debug('Kubectl Cve Hunter determining vulnerable version: {}'.format(self.event.version))
        for vulnerability, fix_versions in cve_mapping.items():
            if CveUtils.is_vulnerable(fix_versions, self.event.version, not config.include_patched_versions):
                self.publish_event(vulnerability(binary_version=self.event.version))



import logging
import threading

from kube_hunter.conf import config
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Event, Service, Vulnerability, HuntFinished, HuntStarted, ReportDispatched


global services_lock
services_lock = threading.Lock()
services = list()

global vulnerabilities_lock
vulnerabilities_lock = threading.Lock()
vulnerabilities = list()

hunters = handler.all_hunters


def console_trim(text, prefix=' '):
    a = text.split(" ")
    b = a[:]
    total_length = 0
    count_of_inserts = 0
    for index, value in enumerate(a):
        if (total_length + (len(value) + len(prefix))) >= 80:
            b.insert(index + count_of_inserts, '\n')
            count_of_inserts += 1
            total_length = 0
        else:
            total_length += len(value) + len(prefix)
    return '\n'.join([prefix + line.strip(' ') for line in ' '.join(b).split('\n')])


def wrap_last_line(text, prefix='| ', suffix='|_'):
    lines = text.split('\n')
    lines[-1] = lines[-1].replace(prefix, suffix, 1)
    return '\n'.join(lines)


@handler.subscribe(Service)
@handler.subscribe(Vulnerability)
class Collector(object):
    def __init__(self, event=None):
        self.event = event

    def execute(self):
        """function is called only when collecting data"""
        global services
        global vulnerabilities
        bases = self.event.__class__.__mro__
        if Service in bases:
            with services_lock:
                services.append(self.event)
            import datetime
            logging.info("|\n| {name}:\n|   type: open service\n|   service: {name}\n|_  location: {location}".format(
                name=self.event.get_name(),
                location=self.event.location(),
                time=datetime.time()
            ))
        elif Vulnerability in bases:
            with vulnerabilities_lock:
                vulnerabilities.append(self.event)
            logging.info(
                "|\n| {name}:\n|   type: vulnerability\n|   location: {location}\n|   description: \n{desc}".format(
                    name=self.event.get_name(),
                    location=self.event.location(),
                    desc=wrap_last_line(console_trim(self.event.explain(), '|     '))
                ))


class TablesPrinted(Event):
    pass


@handler.subscribe(HuntFinished)
class SendFullReport(object):
    def __init__(self, event):
        self.event = event

    def execute(self):
        report = config.reporter.get_report(statistics=config.statistics, mapping=config.mapping)
        config.dispatcher.dispatch(report)
        handler.publish_event(ReportDispatched())
        handler.publish_event(TablesPrinted())


@handler.subscribe(HuntStarted)
class StartedInfo(object):
    def __init__(self, event):
        self.event = event

    def execute(self):
        logging.info("~ Started")
        logging.info("~ Discovering Open Kubernetes Services...")



import json
import logging
from enum import Enum

import re
import requests
import urllib3

from kube_hunter.conf import config
from kube_hunter.core.events import handler
from kube_hunter.core.events.types import Vulnerability, Event, K8sVersionDisclosure
from kube_hunter.core.types import Hunter, ActiveHunter, KubernetesCluster, Kubelet, InformationDisclosure, RemoteCodeExec, AccessRisk
from kube_hunter.modules.discovery.kubelet import ReadOnlyKubeletEvent, SecureKubeletEvent

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


""" Vulnerabilities """
class ExposedPodsHandler(Vulnerability, Event):
    """An attacker could view sensitive information about pods that are bound to a Node using the /pods endpoint"""
    def __init__(self, pods):
        Vulnerability.__init__(self, Kubelet, "Exposed Pods", category=InformationDisclosure)
        self.pods = pods
        self.evidence = "count: {}".format(len(self.pods))


class AnonymousAuthEnabled(Vulnerability, Event):
    """The kubelet is misconfigured, potentially allowing secure access to all requests on the kubelet, without the need to authenticate"""
    def __init__(self):
        Vulnerability.__init__(self, Kubelet, "Anonymous Authentication", category=RemoteCodeExec, vid="KHV036")


class ExposedContainerLogsHandler(Vulnerability, Event):
    """Output logs from a running container are using the exposed /containerLogs endpoint"""
    def __init__(self):
        Vulnerability.__init__(self, Kubelet, "Exposed Container Logs", category=InformationDisclosure, vid="KHV037")


class ExposedRunningPodsHandler(Vulnerability, Event):
    """Outputs a list of currently running pods, and some of their metadata, which can reveal sensitive information"""
    def __init__(self, count):
        Vulnerability.__init__(self, Kubelet, "Exposed Running Pods", category=InformationDisclosure, vid="KHV038")
        self.count = count
        self.evidence = "{} running pods".format(self.count)


class ExposedExecHandler(Vulnerability, Event):
    """An attacker could run arbitrary commands on a container"""
    def __init__(self):
        Vulnerability.__init__(self, Kubelet, "Exposed Exec On Container", category=RemoteCodeExec, vid="KHV039")


class ExposedRunHandler(Vulnerability, Event):
    """An attacker could run an arbitrary command inside a container"""
    def __init__(self):
        Vulnerability.__init__(self, Kubelet, "Exposed Run Inside Container", category=RemoteCodeExec, vid="KHV040")


class ExposedPortForwardHandler(Vulnerability, Event):
    """An attacker could set port forwarding rule on a pod"""
    def __init__(self):
        Vulnerability.__init__(self, Kubelet, "Exposed Port Forward", category=RemoteCodeExec, vid="KHV041")


class ExposedAttachHandler(Vulnerability, Event):
    """Opens a websocket that could enable an attacker to attach to a running container"""
    def __init__(self):
        Vulnerability.__init__(self, Kubelet, "Exposed Attaching To Container", category=RemoteCodeExec, vid="KHV042")


class ExposedHealthzHandler(Vulnerability, Event):
    """By accessing the open /healthz handler, an attacker could get the cluster health state without authenticating"""
    def __init__(self, status):
        Vulnerability.__init__(self, Kubelet, "Cluster Health Disclosure", category=InformationDisclosure, vid="KHV043")
        self.status = status
        self.evidence = "status: {}".format(self.status)


class PrivilegedContainers(Vulnerability, Event):
    """A Privileged container exist on a node. could expose the node/cluster to unwanted root operations"""
    def __init__(self, containers):
        Vulnerability.__init__(self, KubernetesCluster, "Privileged Container", category=AccessRisk, vid="KHV044")
        self.containers = containers
        self.evidence = "pod: {}, container: {}, count: {}".format(containers[0][0], containers[0][1], len(containers))


class ExposedSystemLogs(Vulnerability, Event):
    """System logs are exposed from the /logs endpoint on the kubelet"""
    def __init__(self):
        Vulnerability.__init__(self, Kubelet, "Exposed System Logs", category=InformationDisclosure, vid="KHV045")


class ExposedKubeletCmdline(Vulnerability, Event):
    """Commandline flags that were passed to the kubelet can be obtained from the pprof endpoints"""
    def __init__(self, cmdline):
        Vulnerability.__init__(self, Kubelet, "Exposed Kubelet Cmdline", category=InformationDisclosure, vid="KHV046")
        self.cmdline = cmdline
        self.evidence = "cmdline: {}".format(self.cmdline)


""" Enum containing all of the kubelet handlers """
class KubeletHandlers(Enum):
    PODS = "pods"                                                                                 # GET
    CONTAINERLOGS = "containerLogs/{pod_namespace}/{pod_id}/{container_name}"                        # GET
    RUNNINGPODS = "runningpods"                                                                   # GET
    EXEC = "exec/{pod_namespace}/{pod_id}/{container_name}?command={cmd}&input=1&output=1&tty=1"     # GET -> WebSocket
    RUN = "run/{pod_namespace}/{pod_id}/{container_name}?cmd={cmd}"                                  # POST, For legacy reasons, it uses different query param than exec.
    PORTFORWARD = "portForward/{pod_namespace}/{pod_id}?port={port}"                                # GET/POST
    ATTACH = "attach/{pod_namespace}/{pod_id}/{container_name}?command={cmd}&input=1&output=1&tty=1" # GET -> WebSocket
    LOGS = "logs/{path}"                                                                          # GET
    PPROF_CMDLINE = "debug/pprof/cmdline"                                                         # GET


""" dividing ports for seperate hunters """
@handler.subscribe(ReadOnlyKubeletEvent)
class ReadOnlyKubeletPortHunter(Hunter):
    """Kubelet Readonly Ports Hunter
    Hunts specific endpoints on open ports in the readonly Kubelet server
    """
    def __init__(self, event):
        self.event = event
        self.path = "http://{}:{}/".format(self.event.host, self.event.port)
        self.pods_endpoint_data = ""

    def get_k8s_version(self):
        logging.debug("Passive hunter is attempting to find kubernetes version")
        metrics = requests.get(self.path + "metrics").text
        for line in metrics.split("\n"):
            if line.startswith("kubernetes_build_info"):
                for info in line[line.find('{') + 1: line.find('}')].split(','):
                    k, v = info.split("=")
                    if k == "gitVersion":
                        return v.strip("\"")

    # returns list of tuples of Privileged container and their pod.
    def find_privileged_containers(self):
        logging.debug("Passive hunter is attempting to find privileged containers and their pods")
        privileged_containers = list()
        if self.pods_endpoint_data:
            for pod in self.pods_endpoint_data["items"]:
                for container in pod["spec"]["containers"]:
                    if "securityContext" in container and "privileged" in container["securityContext"] and container["securityContext"]["privileged"]:
                        privileged_containers.append((pod["metadata"]["name"], container["name"]))
        return privileged_containers if len(privileged_containers) > 0 else None

    def get_pods_endpoint(self):
        logging.debug("Attempting to find pods endpoints")
        response = requests.get(self.path + "pods")
        if "items" in response.text:
            return json.loads(response.text)

    def check_healthz_endpoint(self):
        r = requests.get(self.path + "healthz", verify=False)
        return r.text if r.status_code == 200 else False

    def execute(self):
        self.pods_endpoint_data = self.get_pods_endpoint()
        k8s_version = self.get_k8s_version()
        privileged_containers = self.find_privileged_containers()
        healthz = self.check_healthz_endpoint()
        if k8s_version:
            self.publish_event(K8sVersionDisclosure(version=k8s_version, from_endpoint="/metrics", extra_info="on the Kubelet"))
        if privileged_containers:
            self.publish_event(PrivilegedContainers(containers=privileged_containers))
        if healthz:
            self.publish_event(ExposedHealthzHandler(status=healthz))
        if self.pods_endpoint_data:
            self.publish_event(ExposedPodsHandler(pods=self.pods_endpoint_data["items"]))


@handler.subscribe(SecureKubeletEvent)
class SecureKubeletPortHunter(Hunter):
    """Kubelet Secure Ports Hunter
    Hunts specific endpoints on an open secured Kubelet
    """
    class DebugHandlers(object):
        """ all methods will return the handler name if successful """
        def __init__(self, path, pod, session=None):
            self.path = path
            self.session = session if session else requests.Session()
            self.pod = pod

        # outputs logs from a specific container
        def test_container_logs(self):
            logs_url = self.path + KubeletHandlers.CONTAINERLOGS.value.format(
                pod_namespace=self.pod["namespace"],
                pod_id=self.pod["name"],
                container_name=self.pod["container"]
            )
            return self.session.get(logs_url, verify=False).status_code == 200

        # need further investigation on websockets protocol for further implementation
        def test_exec_container(self):
            # opens a stream to connect to using a web socket
            headers={"X-Stream-Protocol-Version": "v2.channel.k8s.io"}
            exec_url = self.path + KubeletHandlers.EXEC.value.format(
                pod_namespace=self.pod["namespace"],
                pod_id=self.pod["name"],
                container_name=self.pod["container"],
                cmd = ""
            )
            return "/cri/exec/" in self.session.get(exec_url, headers=headers, allow_redirects=False ,verify=False).text

        # need further investigation on websockets protocol for further implementation
        def test_port_forward(self):
            headers = {
                "Upgrade": "websocket",
                "Connection": "Upgrade",
                "Sec-Websocket-Key": "s",
                "Sec-Websocket-Version": "13",
                "Sec-Websocket-Protocol": "SPDY"

            }
            pf_url = self.path + KubeletHandlers.PORTFORWARD.value.format(
                pod_namespace=self.pod["namespace"],
                pod_id=self.pod["name"],
                port=80
            )
            self.session.get(pf_url, headers=headers, verify=False, stream=True).status_code == 200
            #TODO: what to return?

        # executes one command and returns output
        def test_run_container(self):
            run_url = self.path + KubeletHandlers.RUN.value.format(
                pod_namespace='test',
                pod_id='test',
                container_name='test',
                cmd = ""
            )
            # if we get a Method Not Allowed, we know we passed Authentication and Authorization.
            return self.session.get(run_url, verify=False).status_code == 405

        # returns list of currently running pods
        def test_running_pods(self):
            pods_url = self.path + KubeletHandlers.RUNNINGPODS.value
            r = self.session.get(pods_url, verify=False)
            return json.loads(r.text) if r.status_code == 200 else False

        # need further investigation on the differences between attach and exec
        def test_attach_container(self):
            # headers={"X-Stream-Protocol-Version": "v2.channel.k8s.io"}
            attach_url = self.path + KubeletHandlers.ATTACH.value.format(
                pod_namespace=self.pod["namespace"],
                pod_id=self.pod["name"],
                container_name=self.pod["container"],
                cmd = ""
            )
            return "/cri/attach/" in self.session.get(attach_url, allow_redirects=False ,verify=False).text

        # checks access to logs endpoint
        def test_logs_endpoint(self):
            logs_url = self.session.get(self.path + KubeletHandlers.LOGS.value.format(
                path=""
            )).text
            return "<pre>" in logs_url

        # returns the cmd line used to run the kubelet
        def test_pprof_cmdline(self):
            cmd = self.session.get(self.path + KubeletHandlers.PPROF_CMDLINE.value, verify=False)
            return cmd.text if cmd.status_code == 200 else None


    def __init__(self, event):
        self.event = event
        self.session = requests.Session()
        if self.event.secure:
            self.session.headers.update({"Authorization": "Bearer {}".format(self.event.auth_token)})
            # self.session.cert = self.event.client_cert
        # copy session to event
        self.event.session = self.session
        self.path = "https://{}:{}/".format(self.event.host, 10250)
        self.kubehunter_pod = {"name": "kube-hunter", "namespace": "default", "container": "kube-hunter"}
        self.pods_endpoint_data = ""

    def get_pods_endpoint(self):
        response = self.session.get(self.path + "pods", verify=False)
        if "items" in response.text:
            return json.loads(response.text)

    def check_healthz_endpoint(self):
        r = requests.get(self.path + "healthz", verify=False)
        return r.text if r.status_code == 200 else False

    def execute(self):
        if self.event.anonymous_auth:
            self.publish_event(AnonymousAuthEnabled())

        self.pods_endpoint_data = self.get_pods_endpoint()
        healthz = self.check_healthz_endpoint()
        if self.pods_endpoint_data:
            self.publish_event(ExposedPodsHandler(pods=self.pods_endpoint_data["items"]))
        if healthz:
            self.publish_event(ExposedHealthzHandler(status=healthz))
        self.test_handlers()

    def test_handlers(self):
        # if kube-hunter runs in a pod, we test with kube-hunter's pod
        pod = self.kubehunter_pod if config.pod else self.get_random_pod()
        if pod:
            debug_handlers = self.DebugHandlers(self.path, pod=pod, session=self.session)
            try:
                # TODO: use named expressions, introduced in python3.8
                running_pods = debug_handlers.test_running_pods()
                if running_pods:
                    self.publish_event(ExposedRunningPodsHandler(count=len(running_pods["items"])))
                cmdline = debug_handlers.test_pprof_cmdline()
                if cmdline:
                    self.publish_event(ExposedKubeletCmdline(cmdline=cmdline))
                if debug_handlers.test_container_logs():
                    self.publish_event(ExposedContainerLogsHandler())
                if debug_handlers.test_exec_container():
                    self.publish_event(ExposedExecHandler())
                if debug_handlers.test_run_container():
                    self.publish_event(ExposedRunHandler())
                if debug_handlers.test_port_forward():
                    self.publish_event(ExposedPortForwardHandler()) # not implemented
                if debug_handlers.test_attach_container():
                    self.publish_event(ExposedAttachHandler())
                if debug_handlers.test_logs_endpoint():
                    self.publish_event(ExposedSystemLogs())
            except Exception as ex:
                logging.debug(str(ex))
        else:
            pass # no pod to check on.

    # trying to get a pod from default namespace, if doesn't exist, gets a kube-system one
    def get_random_pod(self):
        if self.pods_endpoint_data:
            pods_data = self.pods_endpoint_data["items"]
            # filter running kubesystem pod
            is_default_pod = lambda pod: pod["metadata"]["namespace"] == "default" and pod["status"]["phase"] == "Running"
            is_kubesystem_pod = lambda pod: pod["metadata"]["namespace"] == "kube-system" and pod["status"]["phase"] == "Running"
            pod_data = next((pod_data for pod_data in pods_data if is_default_pod(pod_data)), None)
            if not pod_data:
                pod_data = next((pod_data for pod_data in pods_data if is_kubesystem_pod(pod_data)), None)

            if pod_data:
                container_data = next((container_data for container_data in pod_data["spec"]["containers"]), None)
                if container_data:
                    return {
                        "name": pod_data["metadata"]["name"],
                        "container": container_data["name"],
                        "namespace": pod_data["metadata"]["namespace"]
                    }

@handler.subscribe(ExposedRunHandler)
class ProveRunHandler(ActiveHunter):
    """Kubelet Run Hunter
    Executes uname inside of a random container
    """
    def __init__(self, event):
        self.event = event
        self.base_path = "https://{host}:{port}/".format(host=self.event.host, port=self.event.port)

    def run(self, command, container):
        run_url = KubeletHandlers.RUN.value.format(
            pod_namespace=container["namespace"],
            pod_id=container["pod"],
            container_name=container["name"],
            cmd=command
        )
        return self.event.session.post(self.base_path + run_url, verify=False).text

    def execute(self):
        pods_raw = self.event.session.get(self.base_path + KubeletHandlers.PODS.value, verify=False).text
        if "items" in pods_raw:
            pods_data = json.loads(pods_raw)['items']
            for pod_data in pods_data:
                container_data = next((container_data for container_data in pod_data["spec"]["containers"]), None)
                if container_data:
                    output = self.run("uname -a", container={
                        "namespace": pod_data["metadata"]["namespace"],
                        "pod": pod_data["metadata"]["name"],
                        "name": container_data["name"]
                    })
                    if output and "exited with" not in output:
                        self.event.evidence = "uname -a: " + output
                        break

@handler.subscribe(ExposedContainerLogsHandler)
class ProveContainerLogsHandler(ActiveHunter):
    """Kubelet Container Logs Hunter
    Retrieves logs from a random container
    """
    def __init__(self, event):
        self.event = event
        protocol = "https" if self.event.port == 10250 else "http"
        self.base_url = "{protocol}://{host}:{port}/".format(protocol=protocol, host=self.event.host, port=self.event.port)

    def execute(self):
        pods_raw = self.event.session.get(self.base_url + KubeletHandlers.PODS.value, verify=False).text
        if "items" in pods_raw:
            pods_data = json.loads(pods_raw)['items']
            for pod_data in pods_data:
                container_data = next((container_data for container_data in pod_data["spec"]["containers"]), None)
                if container_data:
                    output = requests.get(self.base_url + KubeletHandlers.CONTAINERLOGS.value.format(
                        pod_namespace=pod_data["metadata"]["namespace"],
                        pod_id=pod_data["metadata"]["name"],
                        container_name=container_data["name"]
                    ), verify=False)
                    if output.status_code == 200 and output.text:
                        self.event.evidence = "{}: {}".format(
                            container_data["name"],
                            output.text.encode('utf-8')
                        )
                        return

@handler.subscribe(ExposedSystemLogs)
class ProveSystemLogs(ActiveHunter):
    """Kubelet System Logs Hunter
    Retrieves commands from host's system audit
    """
    def __init__(self, event):
        self.event = event
        self.base_url = "https://{host}:{port}/".format(host=self.event.host, port=self.event.port)

    def execute(self):
        audit_logs = self.event.session.get(self.base_url + KubeletHandlers.LOGS.value.format(
            path="audit/audit.log"
        ), verify=False).text
        logging.debug("accessed audit log of host: {}".format(audit_logs[:10]))
        # iterating over proctitles and converting them into readable strings
        proctitles = list()
        for proctitle in re.findall(r"proctitle=(\w+)", audit_logs):
            proctitles.append(bytes.fromhex(proctitle).decode('utf-8').replace("\x00", " "))
        self.event.proctitles = proctitles
        self.event.evidence = "audit log: {}".format('; '.join(proctitles))
